/*
** Beispiel: Erzeuger/Verbraucher-Problem als Monitor mit Java realisiert.
**
** Uebersetzung des Programms: 
**     javac ErzVerb.java
**
** Start des Programms:
**     java ErzVerb
**
** Das Programm ist nur fuer einen Erzeuger und einen Verbraucher korrekt.
** Ueberpruefen Sie das, indem Sie noch (mindestens) einen Erzeuger- und
** Verbraucher-Thread starten. Ersetzen Sie dann die beiden 'notify()' 
** Aufrufe durch 'notifyAll()' und testen Sie erneut!
*/

/*
**************************************************************************
** Realisierung des beschraenkten Puffers durch ein Array (Ringpuffer)
**  - wr: Schreibposition
**  - rd: Leseposition
*/

class Buffer {
	private int[] buf;
	private int wr;
	private int rd;
	private int size;
	
	public Buffer(int size)
	{
		buf = new int[size];
		wr = 0;
		rd = 0;
		this.size = size;
	}
		
	public void insertItem(int item)
	{
		buf[wr] = item;
		wr = (wr + 1) % size;
	}
	
	public int removeItem()
	{
		int res = buf[rd];
		rd = (rd + 1) % size;
		return res;
	}

	public int getSize()
	{
		return size;
	}
}

/*
**************************************************************************
** Klassen fuer die Erzeuger- und Verbraucher-Threads
*/

class Producer extends Thread
{
	private ErzVerb ev;

	public Producer(ErzVerb ev)
	{
		this.ev = ev;
	}

	public void run()
	{
		int item = 0;
		int i;
		for (i=0; i<1000; i++) {
			item++;               // Erzeuge item ...
			ev.Insert(item);
		}
	}
}

class Consumer extends Thread
{
	private ErzVerb ev;

	public Consumer(ErzVerb ev)
	{
		this.ev = ev;
	}

	public void run()
	{
		int item;
		int i;
		for (i=0; i<1000; i++) {
			item = ev.Remove();
			System.out.println(item);  // Verbrauche item ...
		}
	}
}

public class ErzVerb {

	/*
    **************************************************************************
	** Monitor fuer Erzeuger/Verbraucher-Problem
	*/
	
	/*
	** Variable count zaehlt belegte Pufferplaetze
	*/
	private int count;

	Buffer buffer;

	public ErzVerb(int size)
	{
		buffer = new Buffer(size);
		count = 0;
	}
		
	/*
	** Einfuegeprozedur
	*/
	public synchronized void Insert(int item)
	{
		while (count == buffer.getSize()) {
			try {
				wait();
			}
			catch (InterruptedException ex) {
			}
		}
		buffer.insertItem(item);
		count++;
		if (count == 1)
			notify();     // Achtung: hier sollte notifyAll() stehen!
	}

	/*
	** Entfernprozedur
	*/
	public synchronized int Remove()
	{
		int result;
		while (count == 0) {
			try {
				wait();
			}
			catch (InterruptedException ex) {
			}
		}
		result = buffer.removeItem();
		count--;
		if (count == buffer.getSize()-1)
			notify();     // Achtung: hier sollte notifyAll() stehen!
		return result;
	}

	/*
	** Hauptprogramm
	*/
	public static void main(String argv[])
	{
		ErzVerb ev = new ErzVerb(1);

		/*
		** Erzeuge je einen Thread fuer die Prozedured Producer und Consumer
		*/
		Producer prod = new Producer(ev);
		Consumer cons = new Consumer(ev);

		prod.start();
		cons.start();
	}
}

	
